function [z,M]=r1l_cbc_search(I,opt)
%
%   [z,M]=r1l_cbc_search(I,opt)
%
%   Input:
%
%   I:  frequency index set
%   opt:
%       opt.alg [0 | 1 | {2}]
%           - 0:    search for vector with fixed M, M as opt.M given
%           - 1:    search for vector and M, z_s=M_{s-1}
%           - 2:    search for vector and M, z_s\in[0,M_{s-1}]
%
%       opt.cbcorder [{0} | 1 | 'rand'] 
%           only for opt.alg=0 or opt.alg=2            
%           - 0:    search for each component z_s starts from 0
%           - 1:    search for each component z_s starts from z_{s-1}+1,
%                   result is an ordered vector [usefull for equalweighted
%                   hyperbolic crosses or l1balls]
%           - rand: search randomly each component from 1,...,M-1
%                   or 1,...,M_{s-1}
%       opt.Display [ {off} | verbose | final ]
%           - off:      no information is shown
%           - verbose:  displays informations at the moment you find
%                       anything
%           - final:    displays the resulting vector, if exists, and the
%                       computational time
%       opt.reduce [{'Increase'},'Bisection']
%           only for opt.alg=1 and opt.alg=2            
%           - Increase: lattice size reduce searchs for smallest M
%           - Bisection: lattice size reduce searchs for M using bisections
%                       finds not smallest M in general, but faster
%
%        opt.zstart:   lower dimensional vector used as starting point
%
%   Output:
%
%   vector z and number M such that 
%   \Lambda(z,M)={jz/M, j=0:M-1} is reconstructing rank-1 lattice of I
%   and the Fourier matrix A=exp(2*pi*1i*(0:M1-1)'*(double(I)*(z./M)')')
%   has orthogonal columns.
%
% Copyright 2014 (c) Lutz Kaemmerer
%   
% See also lfft, gsfft, gsifft, gs_search


    if(nargin<2)
        opt.Display='off';
        opt.Inc=0;
        opt.alg=2;
        opt.cbcorder=0;
        opt.reduce='Increase';
    else
        if(~(isfield(opt,'cbcorder') && (max(opt.cbcorder)==1 || strcmp(opt.cbcorder,'rand')==1)))
            opt.cbcorder=0;
        end
        if(~(isfield(opt,'Display') && strmatch(opt.Display, char('off', 'verbose', 'final'),'exact')>0))
            opt.Display='off';
        end
        if(~(isfield(opt,'reduce') && strmatch(opt.reduce, char('Bisection', 'Increase'),'exact')>0))
            opt.reduce='Increase';
        end
    end


    switch(opt.alg)
        case 0
            if(isfield(opt, 'M'))
                [z,M,t1]=cbc_search0(I,opt.M,opt);
            else
                error('Algorithm 0 requires input opt.M');
            end
        case 1
            [z,M,t1]=cbc_search1(I,opt);
        case 2
            [z,M,t1]=cbc_search1(I,opt);
    end
    if(strcmp(opt.Display,'final')==1 || strcmp(opt.Display,'verbose')==1)
    strtoshow=sprintf('generating vector z=[%d',z(1));
    for s=2:length(z)
        strtoshow=[strtoshow,',%d'];
    end
    strtoshow=sprintf([strtoshow,']\n'],z(2:length(z)));
    strtoshow=[strtoshow,sprintf('cbc search time: %d',t1)];
    disp(strtoshow);
end

end


function [z,M,t1]=cbc_search0(I,M,opt)

    sI=size(I);
    if(size(I,1)>M)
        error('You have to sample at least size(I,1) times. So choose M>=size(I,1)!');
    end
    I=double(I);
    M=double(M);

    if(isfield(opt,'zstart') && length(opt.zstart)>0)
        if(prod(size(opt.zstart))==1)
            z(1)=opt.zstart;
        elseif(size(opt.zstart,1)>1)
            z=opt.zstart';
        else
            z=opt.zstart;
        end
        Itest=unique(I(:,1:length(z)),'rows');
        if(length(unique(mod(Itest*z',M)))~=size(Itest,1))
            error('The start vector is no reconstructing vector!');
        end
        loopdimstart=length(z)+1;
    else
        if(strcmp(opt.cbcorder,'rand')==1)
            z=randi(M-1,1);
        else
            z=[1];
        end
        loopdimstart=2;
    end
    
    if(strcmp(opt.Display,'verbose')==1)
        for j=1:length(z)
            fprintf('dimension %d, z_%d=%d\n', [j,j,z(j)])
        end
    end
    ticsearch=tic;
    z=double(z);
    for s=loopdimstart:size(I,2)
        Is=unique(I(:,1:s),'rows');
        tmpstart=1;
        switch(opt.cbcorder)
            case 1
                tmpstart=(z(s-1)+1);
        end
        for l=tmpstart:M-1
            if(strcmp(opt.cbcorder,'rand')==1)
                zs=randi(M-1,1);
            else
                zs=l;
            end
            zs=double(zs);
        
            if  length(unique(mod(Is*[z,zs]',M)))==size(Is,1)
               z=[z,zs];
               break;
            end
            if(mod(zs,ceil(sqrt(M)))==0 && strcmp(opt.Display,'verbose')==1 && strcmp(opt.cbcorder,'rand')==0)
                fprintf('dimension %d, z_%d>%d\n', [s,s,l])
            end
        end
        if(strcmp(opt.Display,'verbose')==1)
            fprintf('dimension %d, z_%d=%d\n', [s,s,z(s)])
        end
    end
    t1=toc(ticsearch);
end


function [z,M,t1]=cbc_search1(I,opt)

I=double(I);

M(1)=max(I(:,1))-min(I(:,1))+1;
z(1)=1;
Itmp=unique(I(:,1),'rows');
for j=M(1):-1:size(Itmp)
    if(length(unique(mod(Itmp,j)))==length(Itmp))
        Mtmp=j;
    end
end
M(1)=Mtmp;

ticsearch=tic;
for s=2:size(I,2)
   
    Is=unique(I(:,1:s),'rows');
    S(s)=max(I(:,s))-min(I(:,s))+1;
    Itmp=unique(I(:,s),'rows');
    for j=S(s):-1:size(Itmp)
        if(length(unique(mod(Itmp,j)))==length(Itmp))
            Stmp=j;
        end
    end
    S(s)=Stmp;
    if(strcmp(opt.Display,'verbose')==1)
        fprintf('dimension %d, S_%d=%d\n', [s,s,S(s)])
    end
    P(s)=M(s-1)*S(s);
    switch(opt.cbcorder)
        case 1
            tmpstart=(z(s-1)+1);
        otherwise
            tmpstart=0;
    end
    for l=tmpstart:M(s-1)
        if(strcmp(opt.cbcorder,'rand')==1)
            zs=randi(M(s-1),1);
        elseif opt.alg==1
            zs=M(s-1);
        else
            zs=l;
        end
        if  length(unique(mod(Is*[z,zs]',P(s))))==size(Is,1)
           z=[z,zs];
           break;
        end
        if(mod(zs,ceil(P(s)/sqrt(P(s))))==0 && strcmp(opt.Display,'verbose')==1 && strcmp(opt.cbcorder,'rand')==0)
            fprintf('dimension %d, z_%d>%d\n', [s,s,l])
        end
    end
    if(strcmp(opt.Display,'verbose')==1)
        fprintf('dimension %d, z_%d=%d\n', [s,s,z(s)])
    end
    if(strcmp(opt.Display,'verbose')==1)
        optM.Display='verbose';
    else
        optM.Display='off';
    end
    optM.Prime=0;
    optM.searchmethod=opt.reduce;
    optM.minM=M(s-1);
    M(s)=r1l_lattice_size_reduce(Is,P(s),z,optM);
    z=mod(z,M(s));
    if(strcmp(opt.Display,'verbose')==1)
    	fprintf('dimension %d: M_{s-1}*S=%d, M_s=%d\n',[s,P(s),M(s)])
        fprintf('dimension %d, z_%d mod M_s=%d\n', [s,s,z(s)])
    end
    end
    t1=toc(ticsearch);
end